@@ -82,13 +82,13 @@ class FirewallZone(object):
         z = self._fw.check_zone(zone)
         return self._zones[z]
 
-    def _error2warning(self, f, name, *args, **kwargs):
-        # transform errors into warnings
+    def _first_except(self, e, f, name, *args, **kwargs):
         try:
             f(name, *args, **kwargs)
         except FirewallError as error:
-            msg = str(error)
-            log.warning("%s: %s" % (name, msg))
+            if not e:
+                return error
+        return e
 
     def add_zone(self, obj):
         obj.settings = { x : LastUpdatedOrderedDict()
@@ -114,6 +114,7 @@ class FirewallZone(object):
         else:
             transaction = use_transaction
 
+        error = None
         for zone in self.get_zones():
             obj = self._zones[zone]
 
@@ -121,8 +122,8 @@ class FirewallZone(object):
 
             # register icmp block inversion setting but don't apply
             if obj.icmp_block_inversion:
-                self._error2warning(self.add_icmp_block_inversion, obj.name,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_icmp_block_inversion, obj.name,
+                                           use_zone_transaction=zone_transaction)
 
             if len(obj.interfaces) > 0 or len(obj.sources) > 0:
                 obj.applied = True
@@ -131,43 +132,46 @@ class FirewallZone(object):
 
             # load zone in case of missing services, icmptypes etc.
             for args in obj.icmp_blocks:
-                self._error2warning(self.add_icmp_block, obj.name, args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_icmp_block, obj.name, args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.forward_ports:
-                self._error2warning(self.add_forward_port, obj.name, *args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_forward_port, obj.name, *args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.services:
-                self._error2warning(self.add_service, obj.name, args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_service, obj.name, args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.ports:
-                self._error2warning(self.add_port, obj.name, *args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_port, obj.name, *args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.protocols:
-                self._error2warning(self.add_protocol, obj.name, args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_protocol, obj.name, args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.source_ports:
-                self._error2warning(self.add_source_port, obj.name, *args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_source_port, obj.name, *args,
+                                           use_zone_transaction=zone_transaction)
             if obj.masquerade:
-                self._error2warning(self.add_masquerade, obj.name,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_masquerade, obj.name,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.rules:
-                self._error2warning(self.add_rule, obj.name, args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_rule, obj.name, args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.interfaces:
-                self._error2warning(self.add_interface, obj.name, args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_interface, obj.name, args,
+                                           use_zone_transaction=zone_transaction)
             for args in obj.sources:
-                self._error2warning(self.add_source, obj.name, args,
-                                    use_zone_transaction=zone_transaction)
+                error = self._first_except(error, self.add_source, obj.name, args,
+                                           use_zone_transaction=zone_transaction)
             # apply icmp accept/reject rule always
             if obj.applied:
-                self._error2warning(self._icmp_block_inversion, True,
-                                    obj.name, zone_transaction)
+                error = self._first_except(error, self._icmp_block_inversion, True,
+                                           obj.name, zone_transaction)
 
         if use_transaction is None:
             transaction.execute(True)
 
+        if error:
+            raise error
+
     def set_zone_applied(self, zone, applied):
         obj = self._zones[zone]
         obj.applied = applied
@@ -230,15 +234,13 @@ class FirewallZone(object):
 
     # settings
 
-    # generate settings record with sender, timeout, mark
-    def __gen_settings(self, timeout, sender, mark=None):
+    # generate settings record with sender, timeout
+    def __gen_settings(self, timeout, sender):
         ret = {
             "date": time.time(),
             "sender": sender,
             "timeout": timeout,
         }
-        if mark:
-            ret["mark"] = mark
         return ret
 
     def get_settings(self, zone):
@@ -307,9 +309,8 @@ class FirewallZone(object):
                     elif key == "icmp_block_inversion":
                         continue
                     elif key == "forward_ports":
-                        mark = obj.settings["forward_ports"][args]["mark"]
                         self._forward_port(enable, _zone, zone_transaction,
-                                            *args, mark_id=mark)
+                                           *args)
                     elif key == "services":
                         self._service(enable, _zone, args, zone_transaction)
                     elif key == "ports":
@@ -323,12 +324,7 @@ class FirewallZone(object):
                     elif key == "masquerade":
                         self._masquerade(enable, _zone, zone_transaction)
                     elif key == "rules":
-                        if "mark" in obj.settings["rules"][args]:
-                            mark = obj.settings["rules"][args]["mark"]
-                        else:
-                            mark = None
-                        self.__rule(enable, _zone,
-                                    Rich_Rule(rule_str=args), mark,
+                        self.__rule(enable, _zone, Rich_Rule(rule_str=args),
                                     zone_transaction)
                     elif key == "interfaces":
                         self._interface(enable, _zone, args, zone_transaction)
@@ -675,8 +671,8 @@ class FirewallZone(object):
 
         return None
 
-    def __rule(self, enable, zone, rule, mark_id, zone_transaction):
-        self._rule_prepare(enable, zone, rule, mark_id, zone_transaction)
+    def __rule(self, enable, zone, rule, zone_transaction):
+        self._rule_prepare(enable, zone, rule, zone_transaction)
 
     def add_rule(self, zone, rule, timeout=0, sender=None,
                  use_zone_transaction=None):
@@ -695,25 +691,20 @@ class FirewallZone(object):
         else:
             zone_transaction = use_zone_transaction
 
-        if type(rule.element) == Rich_ForwardPort:
-            mark = self._fw.new_mark()
-        else:
-            mark = None
-
         if _obj.applied:
-            self.__rule(True, _zone, rule, mark, zone_transaction)
+            self.__rule(True, _zone, rule, zone_transaction)
 
-        self.__register_rule(_obj, rule_id, mark, timeout, sender)
-        zone_transaction.add_fail(self.__unregister_rule, _obj, rule_id, mark)
+        self.__register_rule(_obj, rule_id, timeout, sender)
+        zone_transaction.add_fail(self.__unregister_rule, _obj, rule_id)
 
         if use_zone_transaction is None:
             zone_transaction.execute(True)
 
         return _zone
 
-    def __register_rule(self, _obj, rule_id, mark, timeout, sender):
+    def __register_rule(self, _obj, rule_id, timeout, sender):
         _obj.settings["rules"][rule_id] = self.__gen_settings(
-            timeout, sender, mark=mark)
+            timeout, sender)
 
     def remove_rule(self, zone, rule,
                     use_zone_transaction=None):
@@ -726,31 +717,24 @@ class FirewallZone(object):
             raise FirewallError(errors.NOT_ENABLED,
                                 "'%s' not in '%s'" % (rule, _zone))
 
-        if "mark" in _obj.settings["rules"][rule_id]:
-            mark = _obj.settings["rules"][rule_id]["mark"]
-        else:
-            mark = None
-
         if use_zone_transaction is None:
             zone_transaction = self.new_zone_transaction(_zone)
         else:
             zone_transaction = use_zone_transaction
 
         if _obj.applied:
-            self.__rule(False, _zone, rule, mark, zone_transaction)
+            self.__rule(False, _zone, rule, zone_transaction)
 
-        zone_transaction.add_post(self.__unregister_rule, _obj, rule_id, mark)
+        zone_transaction.add_post(self.__unregister_rule, _obj, rule_id)
 
         if use_zone_transaction is None:
             zone_transaction.execute(True)
 
         return _zone
 
-    def __unregister_rule(self, _obj, rule_id, mark=None):
+    def __unregister_rule(self, _obj, rule_id):
         if rule_id in _obj.settings["rules"]:
             del _obj.settings["rules"][rule_id]
-        if mark:
-            self._fw.del_mark(mark)
 
     def query_rule(self, zone, rule):
         return self.__rule_id(rule) in self.get_settings(zone)["rules"]
@@ -1216,8 +1200,6 @@ class FirewallZone(object):
                                 "'%s:%s:%s:%s' already in '%s'" % \
                                 (port, protocol, toport, toaddr, _zone))
 
-        mark = self._fw.new_mark()
-
         if use_zone_transaction is None:
             zone_transaction = self.new_zone_transaction(_zone)
         else:
@@ -1225,20 +1207,20 @@ class FirewallZone(object):
 
         if _obj.applied:
             self._forward_port(True, _zone, zone_transaction, port, protocol,
-                                toport, toaddr, mark_id=mark)
+                                toport, toaddr)
 
-        self.__register_forward_port(_obj, forward_id, timeout, sender, mark)
+        self.__register_forward_port(_obj, forward_id, timeout, sender)
         zone_transaction.add_fail(self.__unregister_forward_port, _obj,
-                                  forward_id, mark)
+                                  forward_id)
 
         if use_zone_transaction is None:
             zone_transaction.execute(True)
 
         return _zone
 
-    def __register_forward_port(self, _obj, forward_id, timeout, sender, mark):
+    def __register_forward_port(self, _obj, forward_id, timeout, sender):
         _obj.settings["forward_ports"][forward_id] = \
-            self.__gen_settings(timeout, sender, mark=mark)
+            self.__gen_settings(timeout, sender)
 
     def remove_forward_port(self, zone, port, protocol, toport=None,
                             toaddr=None, use_zone_transaction=None):
@@ -1252,8 +1234,6 @@ class FirewallZone(object):
                                 "'%s:%s:%s:%s' not in '%s'" % \
                                 (port, protocol, toport, toaddr, _zone))
 
-        mark = _obj.settings["forward_ports"][forward_id]["mark"]
-
         if use_zone_transaction is None:
             zone_transaction = self.new_zone_transaction(_zone)
         else:
@@ -1261,20 +1241,19 @@ class FirewallZone(object):
 
         if _obj.applied:
             self._forward_port(False, _zone, zone_transaction, port, protocol,
-                                toport, toaddr, mark_id=mark)
+                                toport, toaddr)
 
         zone_transaction.add_post(self.__unregister_forward_port, _obj,
-                                  forward_id, mark)
+                                  forward_id)
 
         if use_zone_transaction is None:
             zone_transaction.execute(True)
 
         return _zone
 
-    def __unregister_forward_port(self, _obj, forward_id, mark):
+    def __unregister_forward_port(self, _obj, forward_id):
         if forward_id in _obj.settings["forward_ports"]:
             del _obj.settings["forward_ports"][forward_id]
-        self._fw.del_mark(mark)
 
     def query_forward_port(self, zone, port, protocol, toport=None,
                            toaddr=None):
@@ -1514,8 +1493,7 @@ class FirewallZone(object):
                         zone_transaction.add_chain(table, chain)
 
                     rules = backend.build_zone_source_interface_rules(enable,
-                                        zone, self._zones[zone].target,
-                                        interface, table, chain, append)
+                                        zone, interface, table, chain, append)
                     zone_transaction.add_rules(backend, rules)
 
     # IPSETS
@@ -1555,11 +1533,10 @@ class FirewallZone(object):
                         zone_transaction.add_chain(table, chain)
 
                     rules = backend.build_zone_source_address_rules(enable, zone,
-                                    self._zones[zone].target, source, table,
-                                    chain)
+                                                        source, table, chain)
                     zone_transaction.add_rules(backend, rules)
 
-    def _rule_prepare(self, enable, zone, rule, mark_id, zone_transaction):
+    def _rule_prepare(self, enable, zone, rule, zone_transaction):
         if rule.family is not None:
             ipvs = [ rule.family ]
         else:
@@ -1584,17 +1561,18 @@ class FirewallZone(object):
             if type(rule.element) == Rich_Service:
                 svc = self._fw.service.get_service(rule.element.name)
 
-                destinations = [rule.destination] if rule.destination else [None]
-
+                destinations = []
                 if len(svc.destination) > 0:
                     if rule.destination:
                         # we can not use two destinations at the same time
                         raise FirewallError(errors.INVALID_RULE,
                                             "Destination conflict with service.")
-                    destinations = []
                     for ipv in ipvs:
                         if ipv in svc.destination and backend.is_ipv_supported(ipv):
                             destinations.append(svc.destination[ipv])
+                else:
+                    # dummy for the following for loop
+                    destinations.append(None)
 
                 for destination in destinations:
                     if enable:
@@ -1714,16 +1692,12 @@ class FirewallZone(object):
                     if toaddr and enable:
                         zone_transaction.add_post(enable_ip_forwarding, ipv)
 
-                filter_chain = "INPUT" if not toaddr else "FORWARD_IN"
-
                 if enable:
-                    zone_transaction.add_chain("mangle", "PREROUTING")
                     zone_transaction.add_chain("nat", "PREROUTING")
-                    zone_transaction.add_chain("filter", filter_chain)
 
                 rules = backend.build_zone_forward_port_rules(
-                                    enable, zone, filter_chain, port, protocol, toport,
-                                    toaddr, mark_id, rule)
+                                    enable, zone, port, protocol, toport,
+                                    toaddr, rule)
                 zone_transaction.add_rules(backend, rules)
 
             # SOURCE PORT
@@ -1782,12 +1756,21 @@ class FirewallZone(object):
             else:
                 raise FirewallError(errors.INVALID_RULE, "Unknown element %s" %
                                     type(rule.element))
-        return mark_id
 
-    def _service(self, enable, zone, service, zone_transaction):
+    def _service(self, enable, zone, service, zone_transaction, included_services=None):
         svc = self._fw.service.get_service(service)
         helpers = self.get_helpers_for_service_modules(svc.modules, enable)
 
+        # First apply any services this service may include
+        if included_services is None:
+            included_services = [service]
+        for include in svc.includes:
+            if include in included_services:
+                continue
+            self.check_service(include)
+            included_services.append(include)
+            self._service(enable, zone, include, zone_transaction, included_services=included_services)
+
         if enable:
             if self._fw.nf_conntrack_helper_setting == 0:
                 zone_transaction.add_chain("raw", "PREROUTING")
@@ -1905,25 +1888,21 @@ class FirewallZone(object):
             zone_transaction.add_rules(backend, rules)
 
     def _forward_port(self, enable, zone, zone_transaction, port, protocol,
-                       toport=None, toaddr=None, mark_id=None):
+                       toport=None, toaddr=None):
         if check_single_address("ipv6", toaddr):
             ipv = "ipv6"
         else:
             ipv = "ipv4"
 
-        filter_chain = "INPUT" if not toaddr else "FORWARD_IN"
-
         if enable:
-            zone_transaction.add_chain("mangle", "PREROUTING")
             zone_transaction.add_chain("nat", "PREROUTING")
-            zone_transaction.add_chain("filter", filter_chain)
 
         if toaddr and enable:
             zone_transaction.add_post(enable_ip_forwarding, ipv)
         backend = self._fw.get_backend_by_ipv(ipv)
         rules = backend.build_zone_forward_port_rules(
-                            enable, zone, filter_chain, port, protocol, toport,
-                            toaddr, mark_id)
+                            enable, zone, port, protocol, toport,
+                            toaddr)
         zone_transaction.add_rules(backend, rules)
 
     def _icmp_block(self, enable, zone, icmp, zone_transaction):
@@ -1966,12 +1945,6 @@ class FirewallZone(object):
         zone_transaction.add_chain("filter", "INPUT")
         zone_transaction.add_chain("filter", "FORWARD_IN")
 
-        # To satisfy nftables backend rule lookup we must execute pending
-        # rules. See nftables.build_zone_icmp_block_inversion_rules()
-        if enable:
-            zone_transaction.execute(enable)
-            zone_transaction.clear()
-
         for backend in self._fw.enabled_backends():
             if not backend.zones_supported:
                 continue
